/*----------------------------------------------------------------------------*/
/* Module:      rexxtags.rex                                                  */
/*                                                                            */
/* Version:     1.1c                                                          */
/*                                                                            */
/* Author:      (c) Jose Maria Blasco <jm@jmblasco.com>                       */
/*                                                                            */
/* This program is part of the REXXTAGS package. The official source for the  */
/* REXXTAGS package is http://www.rextags.org .                               */
/*                                                                            */
/* This software is subject to the terms of the Common Public License. You    */
/* must accept the terms of this license to use this software. Refer to       */
/* the file CPLv1.0.htm included in this package for more information.        */
/*                                                                            */
/* Description: An extension to W. D. Ashley's RSPCOMP.REX Rexx Server Pages  */
/* compiler which accepts a limited form of XML tag support. XML tag          */
/* semantics are implemented using Rexx routines, allowing quick prototyping  */
/* of XML tags in apache. Rexx XML tags can also be used in production        */
/* servers, even if performance can be a little slow. [Updated for v1.1:      */
/* not so slow after the addition of a caching mechanism]                     */ 
/*                                                                            */
/* Acknowledgements: the skeleton for this routine was taken from             */
/* W. D. Ashley's v1.1.1 RSPCOMP.REX of 2002/06/30 from the mod_rexx          */
/* package v 1.2.0.                                                           */
/*                                                                            */
/* Modifications:                                                             */
/*                                                                            */
/* Date       Author      Description                                         */
/* ---------- ----------- --------------------------------------------------- */
/* 2002/10/31 J.M.Blasco  v1.0  Initial REXXTAGS release                      */
/* 2002/12/04 J.M.Blasco  v1.0b Fix: Error in TAGDEPENDENT tags when line..   */
/*                              ..contains a double quote char ('"').         */
/* 2003/05/27 J.M.Blasco  v1.1  Caching support; allow rexx code between..    */
/*                              .. '<%' and '%>'; <%@import%>. statements     */
/* 2003/06/09 J.M.Blasco  v1.1b Add PASSED VARIABLES support, optimize..      */
/*                              ..compiler and generated code, fix bug for..  */
/*                              ..lines > 200 chars, generate <cr> for..      */
/*                              ..INTERPRET under REGINA.                     */
/* 2003/06/15 J.M.Blasco  v1.1c Optimizations, bug fixes, enhanced error..    */
/*                              ..handling, stem support for passed vars,..   */
/*                              ..multi-line support for expressions (not..   */
/*                              ..parameter lists, optional comma), change..  */
/*                              ..passed variables API.                       */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/*                                                                            */
/* Global option variables:                                                   */
/*                                                                            */
/* errmsg   = 0  do not show error messages (default)                         */
/*            1  output error messages to stdout                              */
/*                                                                            */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/*                                                                            */
/* Global state variable:                                                     */
/*                                                                            */
/* state    = 0  parsing HTML lines (starting state)                          */
/*            1  parsing REXX statements delimited by '<?rexx' and '?>' tags  */
/*            2  parsing REXX statements delimited by                         */
/*               '<script type="rexx">' and '</script>' tags                  */
/*                                or                                          */
/*               '<script language="rexx">' and '</script>' tags              */
/*            3  parsing REXX statements delimited by '<%' and '%>' tags      */
/*                                                                            */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/*                                                                            */
/* Return codes:                                                              */
/*                                                                            */
/* 0  successful execution                                                    */
/* 1  no output REXX filename specified                                       */
/* 2  the input RSP file was not found                                        */
/* 3  cannot open the RSP input file                                          */
/* 4  cannot open the REXX output file                                        */
/*                                                                            */
/*----------------------------------------------------------------------------*/

startTime = time('e')

script_version = 'rexxtags_v1.1c' /* Version of this script                   */
maxStringLength = 200             /* Implementation maximum; change as needed */

/*----------------------------------------------------------------------------*/
/* REXXTAGS-enabled directories MUST declare a number of 'taglib' directories */
/* by using mod_rexx 'RexxSetVar' directive (see the docs).                   */
/*----------------------------------------------------------------------------*/

/* This is for manual tests, and for defaults. Change as you like             */
If taglib.0 = 'TAGLIB.0' Then Do
  taglib.0 = 1
  taglib.1 = './'
End

footer_emitted = 0

validchars = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvxyz0123456789'

tagstack.0 = 0 /* To match for end-tags */
parmstack.0 = 0 /* To pass parameters to the end-tag */

/* The following is for this routine internal caching */
tagcache. = 0 /* We don't have cache information by default */

/* Associate an ending to each state */
EndRexxCode.1 = '?>'
EndRexxCode.2 = '</SCRIPT>'
EndRexxCode.3 = '%>'

/* This stem is for (external) caching support */
taglist.  = 0 /* List of tags for cache */

/* This stem is for imports */
import. = 0

lineno = 0 /* For error msgs */

/* set global variables and option defaults */
errmsg = 0
Parse Upper Version rexx_ver

/* get the input filename arguments and options */
cmdline = Arg(1)
If SubStr(cmdline, 1, 1) = '"' Then Parse Var cmdline '"' rspfilename '"' cmdline
Else                                Parse Var cmdline rspfilename cmdline
cmdline = Strip(cmdline)
If SubStr(cmdline, 1, 1) = '"' Then Parse Var cmdline '"' rexfilename '"' option1 .
Else                                Parse Var cmdline     rexfilename     option1 .
If rexfilename = '' Then Do
  Call console_msg 'Error: No output REXX filename specified.'
  Return 1
End
If Translate(option1) = 'ERRMSG' Then errmsg = 1

/* try to open the rsp file */
If rspfilename = '' Then Do
  Call console_msg 'Error: No input RSP File specified.'
  Return 2
End
retc = Stream(rspfilename, 'c', 'open read')
If retc <> 'READY:' Then Do
  Call console_msg 'Error: cannot open file' rspfilename
  Return 3
End

/* open the rex file */
If Pos('OBJREXX', rexx_ver) > 0 Then retc = Stream(rexfilename, 'c', 'open write replace')
Else                                 retc = Stream(rexfilename, 'c', 'open write')

If retc <> 'READY:' Then Do
  Call console_msg 'Error: cannot open file' rexfilename
  Return 4
End
/*----------------------------------------------------------------------------*/
/* Cache support: cache_active tells us wether we are caching or not          */
/*----------------------------------------------------------------------------*/

caching = (cache_active = 1) | (Translate(cache_active) = "YES")

/* The default cache suffix is '.cache.rex'                                   */
If cache_suffix = 'CACHE_SUFFIX' Then cache_suffix = '.cache.rex'

cachefilename = rspfilename||cache_suffix

If caching Then If MayUseCachedCopy() Then Return 0

/* set up an end-of-file indicator on the input rsp file                      */
eof = 0
Call On NotReady Name SetEof

state = 0                            /* initial state is to output HTML lines */
Call REXX_Pgm_Header

/*----------------------------------------------------------------------------*/
/* Main loop: get lines and process them                                      */
/*----------------------------------------------------------------------------*/

Do Forever
  Call GetLine
  If eof <> 0 Then Leave
  Call Process_Line
End

/*----------------------------------------------------------------------------*/
/* If the page contains a </HTML> line, the program emits automatically       */
/* the footer. But if this is a XML page, this is not guaranteed, so we must  */
/* check if the footer was emmitted, and, if not, emit one.                   */
/*----------------------------------------------------------------------------*/

If \footer_emitted Then Call Rexx_Pgm_Footer

/*----------------------------------------------------------------------------*/
/* Close the files.                                                           */
/*----------------------------------------------------------------------------*/
Call Stream rspfilename, 'c', 'close'
Call Stream rexfilename, 'c', 'close'

/*----------------------------------------------------------------------------*/
/* Create or refresh the cache     .                                          */
/*----------------------------------------------------------------------------*/
If caching Then Call CacheCreateOrRefresh

/*----------------------------------------------------------------------------*/
/* Done. Return control to mod_rexx, which will call the generated .rex file. */
/*----------------------------------------------------------------------------*/
Return 0

/*----------------------------------------------------------------------------*/
/* Cache support: cache the XML file.                                         */
/*   We simply copy the resulting REX file, preceding it with a number of     */
/*   timestamps: one for the XML file itself (obviously); one for each        */
/*   TAG used by the file (since if the tag changes, the generated code       */
/*   might also change), and one for each IMPORTed routine.                   */
/*----------------------------------------------------------------------------*/

CacheCreateOrRefresh: Procedure expose rspfilename rexfilename cachefilename taglist. import.
  startTime = Time('e')
  If Stream(rexfilename, 'c', 'open read') <> 'READY:' Then Return
  rc = SysFileDelete(cachefilename)
  If (rc <> 0) & (rc <> 2) Then Return
  Call LineOut cachefilename,  '/*@CACHE FILE 'rspfilename'::'Stream(rspfilename,'c','query timestamp')'::*/'
  Do i = 1 To taglist.0
    Call LineOut cachefilename,'/*@CACHE TAG 'taglist.i'::*/'
  End
  Do i = 1 To import.0
    Call LineOut cachefilename,'/*@CACHE IMPORT 'import.i'::'import.i.0'::*/'
  End
  Call LineOut cachefilename,  '/*@CACHE END */'
  Do While Lines(rexfilename) > 0
    l = LineIn(rexfilename)
    Call LineOut cachefilename,l
  End
  Call LineOut cachefilename,'/* Cached in' (Time('e')-startTime) '*/' 
  Call Stream rexfilename, 'c', 'close'
  Call Stream cachefilename, 'c', 'close'
Return

/*----------------------------------------------------------------------------*/
/* Cache support: Determine whether we can use a cached copy or not.          */
/*   We get the filenames and timestamps stored at the beginning of the       */
/*   cache file and compare the timestamps for each file.                     */ 
/*----------------------------------------------------------------------------*/

MayUseCachedCopy:
  If Stream(cachefilename,'c','query exists') = '' Then Return 0
  ok = 1
  Call Stream cachefilename,'c','open read'
  Do While Lines(cachefilename) > 0
    Parse value LineIn(cachefilename) with '*@CACHE 'verb filename'::'timestamp'::*'
    If verb = 'END' Then Leave
    If Stream(filename,'c','query timestamp') <> timestamp Then Do
      ok = 0
      Leave
    End
  End
  Call Stream cachefilename,'c','close'
  If \ok Then Return 0
  Call LineOut rexfilename,'/* */ Call "'cachefilename'" Arg(1)'
  Call Stream rexfilename,'c','close'
Return 1

/*----------------------------------------------------------------------------*/
/* function which returns the next line, and maintains a line count           */
/*----------------------------------------------------------------------------*/

GetLine: 
  line = LineIn(rspfilename)
  If eof = 1 Then line = ''
  Else lineno = lineno + 1
Return  

/*----------------------------------------------------------------------------*/
/* function which indicates end-of-file                                       */
/*----------------------------------------------------------------------------*/

SetEof:
  eof = 1
Return

/*----------------------------------------------------------------------------*/
/* process a line from the rsp file                                           */
/*----------------------------------------------------------------------------*/

Process_Line:
  uline = Translate(Strip(line))
  
  /* state <> 0: outputing REXX code. See if an end marker is found, else copy code  */
  If state <> 0 Then Do
    If uline = EndRexxCode.state Then state = 0 /* state <- output HTML line(s)      */
    Else Call LineOut rexfilename, line         /* output the REXX pgm line          */
    Return
  End
   
  /* state = 0: See whether there is a start tag for rexx code...                    */
  If uline = '<?REXX' Then state = 1   
  Else If SubStr(uline, 1, 8) = '<SCRIPT ' Then Do
    Parse Var uline '<SCRIPT' . 'TYPE=' lang . '>'
    If lang = '' Then Parse Var uline '<SCRIPT' . 'LANGUAGE=' lang . '>'
    If Strip(Strip(lang,,'"'),,"'") = 'REXX' Then state = 2 
    Else Nop                          /* must be another language (like Javascript)  */
    End
  Else If uline = '<%' Then state = 3
  
  If state <> 0 Then Return     /* Tags must appear alone on the line, so we're done */
	 
  /* State has not changed, process line and output it                               */

  Do Forever
    Call ProcessTags            /* Process tags in the form <[/]prefix:tag [parms][/]> */
    Call ProcessStatements      /* Processing of <%@ statements%>                      */
    Call ProcessExpressions     /* Processing of <%=expressions%>                      */
    If rescan Then Iterate
    Leave
  End

  Call PrintLnStr line                        /* Output the processed HTML line      */

  If Pos('</HTML>', uline) > 0 Then Call REXX_pgm_footer
  /* Doing this separates the mainline code from any REXX  subroutines/functions     */
  /* added at the end of the RSP file, that way we do not just fall into them        */
  /* (a REXX error).                                                                 */

Return

ProcessExpressions:
  rescan = 0 /* Don't rescan unless we get a new line */
  Do While Pos('<%=',line) > 0 
    p1 = Pos('<%=',line)
    p2 = Pos('%>',line,p1+3)
    Do While p1 >= p2
      oldline = Strip(line,'t')
      Call GetLine
      rescan = 1 /* We got a new line, please re-scan in case we get new tags, @-statements, etc */
      If eof == 1 Then Do; 
        Call RexxTagsError "End-of-file found inside an &lt;%=expression%&gt;." 
	line = oldline;
	rescan = 0 /* Don't rescan, it's useless */
	Return 0 
      End
      If Right(oldline,1) = ',' Then                               /* Strip continuation char        */
        oldline = Left(oldline,Length(oldline)-1)
      line = oldline line
      p2 = Pos('%>',line,Length(oldline))
    End
    b1 = Left(line,p1-1)
    expr = SubStr(line,p1+3,p2-p1-3)
    line = SubStr(line,p2+2)
    Call PrintStr b1
    Call Print expr
  End
Return rescan

ProcessStatements:
  Do Forever
    p1 = Pos('<%@',line)
    If p1 = 0 Then Return
    p2 = Pos('%>',line,p1+3)
    If (p1 > p2) Then Return
    b1 = Left(line,p1-1)
    statement = SubStr(line,p1+3,p2-p1-3)
    line = b1||ProcessStatement(statement)||SubStr(line,p2+2)
  End
Return

ProcessStatement: Procedure Expose import. taglib. lineno wwwunparseduri rexfilename
  Parse arg stmt parms
  If stmt = 'import' Then Return ProcessImportStatement(parms)
  Call RexxTagsError "Unrecognized statement '"stmt"'."
Return ""

ProcessImportStatement: Procedure Expose import. taglib. lineno wwwunparseduri rexfilename
  Parse arg parms
  If Words(parms) < 1 Then Do
    Call RexxTagsError "Invalid @IMPORT statement: missing filename."
    Return ""
  End
  Do p = 1 To Words(parms)
    Parse Var parms parm parms
    importfile = ''
    Do i = 1 To taglib.0
      taglib = taglib.i
      If Right(taglib,1) = '"' then taglib = Left(taglib,length(taglib)-1) /* Apache puts extra "s */
      importfile = Stream(taglib||parm,'C','QUERY EXISTS')
      If Length(importfile) = 0 Then
        importfile = Stream(taglib||parm'.rex','C','QUERY EXISTS')
      If Length(importfile) > 0 Then Do
        Do j = 1 To import.0
          If import.j = importfile Then Iterate p
        End
        If Right(Translate(parm),4) = '.REX' Then routineName = Left(parm,Length(parm)-4)
        Else routineName = parm
        im = import.0 + 1
        import.0 = im
        import.im = importfile
        import.im.0 = Stream(importfile,'C','QUERY TIMESTAMP')
        import.im.1 = routineName
        Leave
      End
    End i
    If importfile = '' Then Call RexxTagsError "Cannot import procedure '"parm"'. File not found."
  End p
Return ""

ProcessTags:
  /*-------------------------------------------------------------------------------------------------*/
  /* We now parse the line for xml-like tags (i.e. <[/]prefix:tag[parms][/]>                         */
  /*-------------------------------------------------------------------------------------------------*/

  colonpos = 0  /* Parsing start point */

  /* Avoid parsing lines which obviously have no tags */
  Do parsing = 1 by 1 While Pos(':',line) > 0 & Pos('<',line) > 0
  
    /*-----------------------------------------------------------------------------------------------*/
    /* The following loop exits when:                                                                */
    /*   1.- We found something resembling a tag (may come back later if we can't associate it with  */
    /*       a REXX routine).                                                                        */
    /*   2.- We definitely do not find a tag (and we then leave the whole parsing phase).            */
    /*-----------------------------------------------------------------------------------------------*/
    Do Forever                                                     /* Parse for tags in a line       */
      pos = colonpos + 1                                           /* Where to look for a colon      */
      colonpos = Pos(':',line,pos)                                 /* If we don't find one...        */
    If colonpos == 0 then Leave parsing                            /* ...there's no tag.             */
      ltpos = LastPos('<',line,colonpos-1)                         /* Now look backwards for a '<'.  */
      If ltpos == 0 then Iterate                                   /* No '<'? Look for next ':'      */
      prefix = SubStr(line,ltpos+1,colonpos-ltpos-1)               /* Extract the prefix of the...   */
      pref = prefix                                                /* tag ('<prefix:tag parms>').    */
      If Left(prefix,1) = '/' then pref = SubStr(prefix,2)         /* Skip '/' for end-tags          */
      If Length(pref) == 0 then Iterate                            /* Prefix must not be empty...    */
      If Verify(pref,validchars) <> 0 then Iterate                 /* ...and chars must be valid.    */
      gtpos = Pos('>',line,colonpos+1)                             /* Now look for tag termination.  */
      If gtpos > 1 Then                                            /* Skip in case we have a...   .  */
        Do While gtpos > 1 & SubStr(line,gtpos-1,1) = '%'          /* <%=expression%>             .  */
          gtpos = Pos('>',line,gtpos+1)
          If gtpos = 0 Then Leave
        End
      If gtpos == 0 then Do                                        /* Maybe in a different line?     */
        blankpos = Pos(' ',line' ',colonpos+1)                     /* Extract the...                 */
        tag = SubStr(line,colonpos+1,blankpos-colonpos-1)          /* ...tag name.                   */
        If Verify(tag,validchars) <> 0 then Iterate                /* Chars must be valid            */
        Do while eof == 0 & gtpos == 0                             /* We now read from the source... */
          oldline = line
          Call GetLine                                             /* ...file until we find a '>'... */
          If eof == 1 Then Do; Leave parsing; line = oldline; End  /* ...char OR the file is ended.  */
	  lline = Length(oldline)
	  line = oldline line                                      /* We assemble a new line, ...    */
	  gtpos = Pos('>',line,lline)                              /* ...and reparse for '>'.        */
	  If gtpos > 1 Then Do while gtpos > 1 & SubStr(line,gtpos-1,1) = '%'
            gtpos = Pos('>',line,gtpos+1)
            If gtpos = 0 Then Leave
          End
        End                                                        /* Now we might have the tag...   */
      End                                                          /* ...1 line.                     */
      Parse Value SubStr(line,colonpos+1,gtpos-colonpos-1) with tag parms /* Parse tag and params.   */
      If Length(tag) == 0 then Iterate                             /* Tag may not be empty, ...      */
      If tag = '/' Then Iterate
      If Right(tag,1) = '/' & Length(parms) == 0 Then Do
        If Verify(Left(tag,Length(tag)-1),validchars) <> 0 Then Iterate
      End
    Else If Verify(tag,validchars) <> 0 then Iterate               /* ...and chars must be valid.    */
       Leave                                                       /* Now the tag is parsed          */
    End                                                            

    /*-----------------------------------------------------------------------------------------------*/
    /* We extract now the 'before' and 'after' parts of the line.                                    */
    /*-----------------------------------------------------------------------------------------------*/

    before = Left(line,ltpos-1)
    after = SubStr(line,gtpos+1)

    /*-----------------------------------------------------------------------------------------------*/
    /* Determine the form of the tag. This will influence how the associated REXX routine is called. */
    /*-----------------------------------------------------------------------------------------------*/
	   
    If Left(prefix,1) = '/' then Do                                /* '</prefix:tag[parms]>'         */
      tagtype = 'END'
      prefix = SubStr(prefix,2)
      End
    Else If Length(parms) > 0 & Right(parms,1) = '/' then Do       /* '<prefix:tag parms/>'          */
      tagtype = 'BOTH'
      parms = Left(parms,Length(parms)-1)
      End
    Else If Length(parms) == 0 & Right(tag,1) = '/' then Do        /* '<prefix:tag/>'                */
      tagtype = 'BOTH'
      tag = Left(tag,Length(tag)-1)
      End
    Else Do                                                        /* '<prefix:tag>'                 */
      tagtype = 'START'
    End
	   
    /*-----------------------------------------------------------------------------------------------*/
    /* We have now to find a suitable REXX routine that implements the tag in one of the 'taglibs'.  */
    /*-----------------------------------------------------------------------------------------------*/
	   
    rexxtagfile = tagcache.prefix.tag                              /* Look in the tag cache          */
    If rexxtagfile = '' then                                       /* Tag is unknown. Look for...    */
      Iterate parsing                                              /* ...more tags.                  */
	   
    If rexxtagfile = 0 then Do                                     /* We don't have cache info       */ 
      found = 0
      Do i = 1 to taglib.0
        taglib = taglib.i
        If Right(taglib,1) = '"' then taglib = Left(taglib,length(taglib)-1) /* Apache puts extra "s */
          rexxtagfile = Stream(taglib||prefix'.'tag'.tag','C','QUERY EXISTS')
        If Length(rexxtagfile) > 0 then Do
          tagcache.tag.prefix = rexxtagfile                        /* Cache the REXX file            */
          If taglist.0.rexxtagfile = 0 Then Do
            taglist.0 = taglist.0 + 1
            tl0 = taglist.0
            taglist.0.rexxtagfile = 1
            taglist.tl0 = rexxtagfile'::'Stream(rexxtagfile,'c','query timestamp')
          End
          found = 1
          Leave
        End
      End
      /* FIXME: Should store smth in the cache, in case the tag is created later! */
      If \found then Do                                            /* Tag is unknown.                */
        tagcache.tag.prefix = ''                                   /* Cache the fact, ..             */
	If tagtype = 'END' Then Do
          ii = tagstack.0
          If ii > 0 & tagstack.ii <> prefix':'tag Then 
	    Call RexxTagsError "&lt;/"prefix":"tag"&gt; is no matching end-tag for &lt;"tagstack.ii"&gt; and was discarded." 
	End
        Iterate parsing                                            /* ...and look for more tags.     */
      End
      Interpret 'import = "'rexxtagfile'"("QUERY IMPORT")'
      If import <> 0 Then Do
        r = ProcessImportStatement(import)
        If r <> '' Then Do
          Call RexxTagsError "Import error from &lt;"prefix":"tag"&gt;:" r
        End
      End
    End

    /* Parms are NOT evaluated for %-expressions for this call */
    Interpret 'bodycontent = "'rexxtagfile'"("QUERY BODY CONTENT",parms)'
    	   
    /* Processing of <%=(expr)%> in before */
    /* Because of the way we process the line, we don't need to implement multi-line support here */
    Do While pos('<%=',before) > 0
      p1 = pos('<%=',before)
      p2 = pos('%>',before,p1+3)
      If (p1 > p2) Then Leave
      b1 = Left(before,p1-1)
      expr = Substr(before,p1+3,p2-p1-3)
      If Length(b1) > 0 Then Call PrintStr b1
      Call Print expr
      before = SubStr(before,p2+2)
    End
    If Length(before) > 0 Then Call PrintStr before

    /* expressions in parameters and passed variables */
    If tagtype <> 'END' Then Do
      Interpret 'queryApacheRequest = "'rexxtagfile'"("QUERY APACHE REQUEST")'
      If      queryApacheRequest = 0 Then Call LineOut rexfilename, "___parms = ''"
      Else If queryApacheRequest = 1 Then Call LineOut rexfilename, "___parms = ' request=""'Arg(1)'"" ' "
      Else                                Call LineOut rexfilename, "___parms = '"queryApacheRequest"=""Arg(1)"" ' "
      /* Processing of <%=(expr)%> in parms  */
      /* We DON'T accept multi-line in parms */
      Do While pos('<%=',parms) > 0 
        p1 = pos('<%=',parms)
        p2 = pos('%>',parms,p1+3)
        If (p1 > p2) Then Leave
        b1 = Left(parms,p1-1)
        expr = Substr(parms,p1+3,p2-p1-3)
        parms = SubStr(parms,p2+2)
	If Length(b1) > 0 Then Call PrintParmStr b1
	Call PrintParmExpr expr
      End
      If Length(parms) > 0 Then Call PrintParmStr parms
      /* Processing of passed variables  */
      Call LineOut rexfilename, "___vars = ___passed_variables( '"rexxtagfile"'('"VARIABLES"',___parms) )"
    End

    /*-----------------------------------------------------------------------------------------------*/
    /* We found the associated REXX routine. Call it as appropriate, and, if necessary, stack tag    */
    /* and parameters for end-tag processing and matching                                            */
    /*-----------------------------------------------------------------------------------------------*/

    currentTag = prefix':'tag
    
    Select
      When Translate(bodyContent) = 'TAGDEPENDENT' Then Call ProcessTagDependentTag
      When (tagtype == 'END') & (TagStackTop()) <> currentTag Then 
        Call RexxTagsError "&lt;/"currentTag"&gt; is no matching end-tag for &lt;"TagStackTop()"&gt; and is ignored."
      When tagtype = 'START' & (Translate(bodycontent) <> 'RSP') Then Do
        Call Push_Tag         currentTag
	Call Store_Tag_Mode  "NORMAL"
	Call Push_State
	Call Make_Start_Call
        End /* tagtype = 'START', no RSP */
      When (tagtype = 'START') & (Translate(bodycontent) = 'RSP') then Do
        Call Push_Tag             currentTag
	Call Store_Tag_Mode      "RSP"
	Call Push_State
	Call Make_Start_Call
        Call Start_Body
        End /* tagtype = 'START', RSP */
      When tagtype = 'END' then Do
        tagMode = RetrieveTagMode()
        Call Pop_Tag
        Call Pop_State
        If tagMode == 'RSP' Then Do
	  Call End_Body
	  Call Make_Body_Call
        End
	Call Make_End_Call
        End /* tagtype = 'END' */
      When (tagtype = 'BOTH') & (Translate(bodycontent) <> 'RSP') then Do
        Call Make_Start_Call 
	Call Make_End_Call
        End /* tagtype = 'BOTH', no RSP */
      When (tagtype = 'BOTH') & (Translate(bodycontent) = 'RSP') then Do
 	Call Make_Start_Call
        Call Start_Body
        Call End_Body
	Call Make_Body_Call
 	Call Make_End_Call
        End /* tagtype = 'BOTH', RSP */
    End /* Select */
    line = after
    colonpos = 0
  End /* parsing */
Return

ProcessTagDependentTag:
  Call Make_Start_Call
  Call Start_Body
  If tagType <> 'BOTH' Then Do
    endTag = '</'prefix':'tag'>'
    Do Forever
      p =  Pos(endTag,after)
      If p > 0 Then Do
        Call PrintLnStr SubStr(after,1,p-1)
        after = SubStr(after,p+length(endTag))
        Leave
      End
      Call PrintLnStr after
      If eof == 0 Then Call GetLine
      If eof == 1 Then Do
        Call End_Body
	Call Discard_Body
        Call RexxTagsError "No matching end-tag for &lt;"prefix":"tag"&gt;.</b><br>"
        after = ''
        Return
      End
      after = line
    End
  End
  Call End_Body
  Call Make_Body_Call
  Call Make_End_Call
Return

RexxTagsError: /* Uses SAY instead of PrintLn to ensure output */
  Call lineout rexfilename, 'Say "<hr color=""blue""><b><font color=""blue"">REXXTAGS: compilation error in "WWWUNPARSEDURI" at line"' lineno'": 'add_quotes('"',Arg(1))'</font></b><hr color=""blue"">"'
Return

/*----------------------------------------------------------------------------*/
/* PrintLn a string. Support for large lines.                                 */
/*----------------------------------------------------------------------------*/

PrintLnStr: Procedure Expose rexfilename maxStringLength
  line = Arg(1)
  Do While Length(line) > maxStringLength
    l = Left(line,maxStringLength)
    line = SubStr(line,maxStringLength+1)
    Call LineOut rexfilename, "Call Print """DDQ(l)""""
  End
  Call LineOut rexfilename, "Call PrintLn """DDQ(line)""""
Return

/*----------------------------------------------------------------------------*/
/* PrintLn a token                                                            */
/*----------------------------------------------------------------------------*/

PrintLn: Procedure Expose rexfilename maxStringLength
  Call LineOut rexfilename, "Call PrintLn "Arg(1)
Return

/*----------------------------------------------------------------------------*/
/* Print a string. Support for large lines.                                   */
/*----------------------------------------------------------------------------*/

PrintStr: Procedure Expose rexfilename maxStringLength
  line = Arg(1)
  Do While Length(line) > maxStringLength
    l = Left(line,maxStringLength)
    line = SubStr(line,maxStringLength+1)
    Call LineOut rexfilename, "Call Print """DDQ(l)""""
  End
  Call LineOut rexfilename, "Call Print """DDQ(line)""""
Return

Print: Procedure Expose rexfilename maxStringLength 
  Call LineOut rexfilename, "Call Print "Arg(1)
Return

/*----------------------------------------------------------------------------*/
/* Create (part of) the parameter list. Support for large lines               */
/*----------------------------------------------------------------------------*/

PrintParmStr: Procedure Expose rexfilename maxStringLength
  parms = Arg(1)
  Do While length(parms) > maxStringLength
    part = Left(parms,maxStringLength)
    parms = Substr(parms,maxStringLength+1)
    Call LineOut rexfilename, "___parms = ___parms||"""DDQ(part)""""
  End
  Call LineOut rexfilename, "___parms = ___parms||"""DDQ(parms)""""
Return

PrintParmExpr: 
  Call lineout rexfilename, "___parms = ___parms||("Arg(1)")"
Return

/*----------------------------------------------------------------------------*/
/* Compilation-time tag stack primitives (error checking)                    */
/*----------------------------------------------------------------------------*/

Push_Tag: Procedure Expose tagstack.  
  i = tagstack.0 + 1
  tagstack.i = Arg(1)
  tagstack.0 = i
Return

Pop_Tag: Procedure Expose tagstack.
  tagstack.0 = tagstack.0 - 1
Return

Store_Tag_Mode: Procedure Expose tagstack. /* Should always be called after 'Push_Tag' */
  i = tagstack.0
  tagstack.i.0 = Arg(1)
Return

RetrieveTagMode: Procedure Expose tagstack.
 i = tagStack.0
Return tagStack.i.0

TagStackTop: Procedure Expose tagStack.
 i = tagStack.0
Return tagStack.i

/*----------------------------------------------------------------------------*/
/* Execution-time parameter-list stacking primitives                          */
/*----------------------------------------------------------------------------*/
Push_State: 
  Call LineOut rexfilename, "Call ___Push_State"  
Return

Pop_State:  
  Call LineOut rexfilename, "Call ___Pop_State"   
Return

/*----------------------------------------------------------------------------*/
/* Execution-time markers for bodies. (Influences who will get the output)    */
/*----------------------------------------------------------------------------*/
Start_Body:   Call LineOut rexfilename, "Call ___start_body";  Return
End_Body:     Call LineOut rexfilename, "Call ___end_body";    Return
Discard_Body: Call LineOut rexfilename, "Call ___pop_body";    Return

/*----------------------------------------------------------------------------*/
/* The three standard calls                                                   */
/*----------------------------------------------------------------------------*/
Make_Start_Call: Call LineOut rexfilename, "Call ___process """rexxtagfile"""('START',___parms,,___vars)";              Return
Make_Body_Call:  Call LineOut rexfilename, "Call ___process """rexxtagfile"""('BODY', ___parms,___pop_body(),___vars)"; Return
Make_End_Call:   Call LineOut rexfilename, "Call ___process """rexxtagfile"""('END',  ___parms,,___vars)";              Return

/*----------------------------------------------------------------------------*/
/* add quotes to a string (two for one)                                       */
/*----------------------------------------------------------------------------*/

add_quotes: procedure
quote = Arg(1)
line = Arg(2)
x = Pos(quote, line)
Do While x > 0
   line = SubStr(line, 1, x) || quote || SubStr(line, x + 1)
   x = Pos(quote, line, x + 2)
End
Return line

/*----------------------------------------------------------------------------*/
/* add header information to the REXX program                                 */
/*----------------------------------------------------------------------------*/

REXX_pgm_header:
Call LineOut rexfilename, '/* Start of REXX RSP header information */'
Call LineOut rexfilename, '/*'
Call LineOut rexfilename, ''

Call LineOut rexfilename, 'RSP Compiler Information:'
Call LineOut rexfilename, '   Compiler version: ' script_version
Call LineOut rexfilename, '   Compile timestamp:' Date() Time()
Call LineOut rexfilename, '   Operating System: ' os
Call LineOut rexfilename, '   REXX Version:     ' rexx_ver
Call LineOut rexfilename, ''

Call LineOut rexfilename, 'Source RSP File Information:'
fsize      = Stream(rspfilename, 'c', 'query size')
ftimestamp = Stream(rspfilename, 'c', 'query timestamp')
Call LineOut rexfilename, '   ' || ftimestamp || '  ' || Format(fsize, 10) || ,
                          '  ' || Stream(rspfilename, 'c', 'query exists')
Call LineOut rexfilename, ''

Call LineOut rexfilename, 'Output REXX Program File Information:'
Call LineOut rexfilename, '   Filename:' Stream(rexfilename, 'c', 'query exists')
Call LineOut rexfilename, ''

Call LineOut rexfilename, '*/'

Call LineOut rexfilename, ''
Call LineOut rexfilename, 'Call WWWSendHTTPHeader arg(1), "text/html"'
Call LineOut rexfilename, ''

Call LineOut rexfilename, "___parmstack.0 = 0"
Call LineOut rexfilename, "___varstack.0 = 0"
Call LineOut rexfilename, "___rest = ''"
Call LineOut rexfilename, "___body.0 = 0"
Call LineOut rexfilename, "___bodyn = 0"
If Pos('REGINA',rexx_ver) > 0 Then Call LineOut rexfilename, "___nl = '0a'x"
Else Call LineOut rexfilename, "___nl = ''"

Call LineOut rexfilename, '/* End of REXX RSP header information */'

Call LineOut rexfilename, ''
Return


/*----------------------------------------------------------------------------*/
/* add footer information to the REXX program                                 */
/*----------------------------------------------------------------------------*/

REXX_pgm_footer:
footer_emitted = 1

If tagstack.0 > 0 Then 
  Do i = tagstack.0 To 1 By -1
    Call RexxTagsError "Tag &lt;"tagstack.i"&gt; not closed."
  End

Call LineOut rexfilename, '/* Start of REXX RSP footer information */'
Call LineOut rexfilename, 'Return 0'

/* Add all IMPORTed routines here                                             */
Do i = 1 To import.0
  Call LineOut rexfilename,"/* IMPORTed routine "import.i.1" */"
  Call LineOut rexfilename,import.i.1':'
  importfile = import.i
  Call Stream importfile,'c','open read'
  Do While Lines(importfile) > 0
    Call LineOut rexfilename,LineIn(importfile)
  End
  Call Stream importfile,'c','close'
End

                          /* Subroutine to collect passed variables */
Call LineOut rexfilename, "___passed_variables: Parse Arg ___r"
Call LineOut rexfilename, "___vars = ''"
Call LineOut rexfilename, "If ___r == 0 Then Return ''"
Call LineOut rexfilename, "If ___r == '' Then Return ''"
Call LineOut rexfilename, "Do ___i = 1 To Words(___r)"
Call LineOut rexfilename, "  Parse Var ___r ___v ___r"
Call LineOut rexfilename, "  If right(___v,1) == '.' Then Do"
Call LineOut rexfilename, "    ___t = Value(___v'0')"
Call LineOut rexfilename, "    If DataType(___t,'w') Then Do"
Call LineOut rexfilename, "      Do ___j = 1 To ___t"
Call LineOut rexfilename, "        ___s = Value(___v||___j)"
Call LineOut rexfilename, "        ___x = ___v||___j '=' '''''';"
Call LineOut rexfilename, "        Do While Length(___s) >" maxStringLength
Call LineOut rexfilename, "          ___x = ___x';'___nl ___v||___j '=' ___v||___j''''___DSQ(Left(___s,"maxStringLength"))''''"
Call LineOut rexfilename, "          ___s = SubStr(___s,"maxStringLength"+1)"
Call LineOut rexfilename, "        End" 
Call LineOut rexfilename, "        ___x = ___x';'___nl ___v||___j '= '___v||___j''''___s''''"
Call LineOut rexfilename, "        If ___vars == '' Then ___vars = ___x"
Call LineOut rexfilename, "        Else ___vars = ___vars';'___nl ___x"
Call LineOut rexfilename, "      End"
Call LineOut rexfilename, "      ___v = ___v'0'"
Call LineOut rexfilename, "    End"
Call LineOut rexfilename, "  End"
Call LineOut rexfilename, "  ___s = Value(___v)"
Call LineOut rexfilename, "  ___x = ___v '=' '''''';"
Call LineOut rexfilename, "  Do While Length(___s) >" maxStringLength
Call LineOut rexfilename, "   ___x = ___x';'___nl ___v '=' ___v'||'''___DSQ(Left(___s,"maxStringLength"))''''"
Call LineOut rexfilename, "   ___s = SubStr(___s,"maxStringLength"+1)"
Call LineOut rexfilename, "  End" 
Call LineOut rexfilename, "  ___x = ___x';'___nl ___v '= '___v'||'''___s''''"
Call LineOut rexfilename, "  If ___vars == '' Then ___vars = ___x"
Call LineOut rexfilename, "  Else ___vars = ___vars';'___nl ___x"
Call LineOut rexfilename, "End"
Call LineOut rexfilename, "Return ___vars"

                          /* Subroutine to process tag results */
Call LineOut rexfilename, "___process: Parse arg ___r"
Call LineOut rexfilename, "___state = 0 /* HTML Code */"
Call LineOut rexfilename, "___firstLine = 1"
Call LineOut rexfilename, "Do while pos('00'x,___r) > 0"
Call LineOut rexfilename, "  Parse var ___r ___s '00'x ___r"
                             /* Start of REXX code. Switch status to REXX */
Call LineOut rexfilename, "  If ___state == 0 & ___s = '<%' Then Do"
Call LineOut rexfilename, "    ___state = 1 /* REXX Code */"
Call LineOut rexfilename, "    ___stmt = ''"
Call LineOut rexfilename, "    End"
                             /* End of REXX code. Interpret the REXX code and switch status back to HTML */
Call LineOut rexfilename, "  Else If ___state == 1 & ___s = '%>' Then Do"
Call LineOut rexfilename, "    Interpret ___stmt"
Call LineOut rexfilename, "    ___stmt = ''"
Call LineOut rexfilename, "    ___state = 0"
Call LineOut rexfilename, "    End"
                             /* That's a REXX line. Collect it into a large statement */
Call LineOut rexfilename, "  Else If ___state == 1 Then Do"
Call LineOut rexfilename, "    If Right(Strip(___stmt),1) = ',' Then ___stmt = Left(Strip(___stmt),Length(Strip(___stmt))-1) ___s"
Call LineOut rexfilename, "    Else ___stmt = ___stmt';'___nl ___s"
Call LineOut rexfilename, "    End"
                             /* That's HTML */
Call LineOut rexfilename, "  Else Do"
Call LineOut rexfilename, "    If ___firstLine Then Do; Call Print ___s;  ___firstLine = 0; End"
Call LineOut rexfilename, "    Else Do; Call PrintLn; Call Print ___s; End"
Call LineOut rexfilename, "  End"
Call LineOut rexfilename, "End"

                          /* We got out of the loop. See whether we must interpret REXX, or... */
Call LineOut rexfilename, "If ___state == 1 & ___r = '%>' Then Interpret ___stmt"
                          /* ...output HTML. */
Call LineOut rexfilename, "Else Do"
Call LineOut rexfilename, "  If ___firstLine Then Call Print ___r; Else Do; Call PrintLn; Call Print ___r; End"
Call LineOut rexfilename, "End"
Call LineOut rexfilename, "Return"

Call LineOut rexfilename, "___start_body:"
Call LineOut rexfilename, "___bodyn = ___body.0 + 1"
Call LineOut rexfilename, "___body.0 = ___body.0 + 1"
Call LineOut rexfilename, "___body.___bodyn = 'body.1 = '''''"
Call LineOut rexfilename, "___bodylines.___bodyn = 1"
Call LineOut rexfilename, "Return"

Call LineOut rexfilename, "___end_body:"
Call LineOut rexfilename, "___body.___bodyn = ___body.___bodyn ||'; body.0 = '___bodylines.___bodyn"
Call LineOut rexfilename, "Return"

Call LineOut rexfilename, "___pop_body:"
Call LineOut rexfilename, "___body.0 = ___body.0 - 1"
Call LineOut rexfilename, "___r = ___body.___bodyn"
Call LineOut rexfilename, "Drop ___body.___bodyn"
Call LineOut rexfilename, "___bodyn = ___body.0"
Call LineOut rexfilename, "Return ___r"

Call LineOut rexfilename, "___Push_State:"
Call LineOut rexfilename, "___i = ___parmstack.0 + 1"
Call LineOut rexfilename, "___parmstack.0 = ___i"
Call LineOut rexfilename, "___parmstack.___i = ___parms"
Call LineOut rexfilename, "___i = ___varstack.0 + 1"
Call LineOut rexfilename, "___varstack.0 = ___i"
Call LineOut rexfilename, "___varstack.___i = ___vars"
Call LineOut rexfilename, "Return"

Call LineOut rexfilename, "___Pop_State:"
Call LineOut rexfilename, "___i = ___parmstack.0"
Call LineOut rexfilename, "___parms = ___parmstack.___i"
Call LineOut rexfilename, "___parmstack.0 = ___i - 1"
Call LineOut rexfilename, "___i = ___varstack.0"
Call LineOut rexfilename, "___vars = ___varstack.___i"
Call LineOut rexfilename, "___varstack.0 = ___i - 1"
Call LineOut rexfilename, "Return"

Call LineOut rexfilename, "PrintLn:"
Call LineOut rexfilename, "If Arg(1) <> '' Then Do"
Call LineOut rexfilename, "  If ___bodyn == 0 Then ___rest = ___rest||Arg(1)"
Call LineOut rexfilename, '  Else ___body.___bodyn = ___body.___bodyn"; body."___bodylines.___bodyn" = body."___bodylines.___bodyn"''"___DSQ(Arg(1))"''"'
Call LineOut rexfilename, "End"
Call LineOut rexfilename, "If ___bodyn == 0 Then Do"
Call LineOut rexfilename, "  Say ___rest"
Call LineOut rexfilename, "  ___rest = ''"
Call LineOut rexfilename, "  End"
Call LineOut rexfilename, "Else Do"
Call LineOut rexfilename, "  ___bodylines.___bodyn = ___bodylines.___bodyn + 1"
Call LineOut rexfilename, "  ___body.___bodyn = ___body.___bodyn||"";""___nl""body.""___bodylines.___bodyn"" = ''"""
Call LineOut rexfilename, "End"
Call LineOut rexfilename, "Return"

Call LineOut rexfilename, "Print:"
Call LineOut rexfilename, "If Arg(1) == '' Then Return"
Call LineOut rexfilename, "If ___bodyn == 0 Then ___rest = ___rest||Arg(1)"
Call LineOut rexfilename, 'Else ___body.___bodyn = ___body.___bodyn"; "___nl"body."___bodylines.___bodyn" = body."___bodylines.___bodyn"''"___DSQ(Arg(1))"''"'
Call LineOut rexfilename, "Return"

Call LineOut rexfilename, "___DSQ: procedure"
Call LineOut rexfilename, "line = arg(1)"
Call LineOut rexfilename, "x = pos(""'"", line)"
Call LineOut rexfilename, "Do while x > 0"
Call LineOut rexfilename, "   line = substr(line, 1, x)""'""substr(line, x + 1)"
Call LineOut rexfilename, "   x = pos(""'"", line, x + 2)"
Call LineOut rexfilename, "   end"
Call LineOut rexfilename, "Return line"

Call LineOut rexfilename, '___DDQ: Procedure'
Call LineOut rexfilename, 'line = arg(1)'
Call LineOut rexfilename, 'x = pos(''"'', line)'
Call LineOut rexfilename, 'Do while x > 0'
Call LineOut rexfilename, '   line = substr(line, 1, x)''"''substr(line, x + 1)'
Call LineOut rexfilename, '   x = pos(''"'', line, x + 2)'
Call LineOut rexfilename, 'End'
Call LineOut rexfilename, 'Return line'

Call LineOut rexfilename, '___QDQ: Procedure'
Call LineOut rexfilename, 'line = Arg(1)'
Call LineOut rexfilename, 'x = Pos(''"'', line)'
Call LineOut rexfilename, 'Do While x > 0'
Call LineOut rexfilename, '   line = SubStr(line, 1, x)''"""''SubStr(line, x + 1)'
Call LineOut rexfilename, '   x = Pos(''"'', line, x + 2)'
Call LineOut rexfilename, 'End'
Call LineOut rexfilename, 'Return line'

Call LineOut rexfilename, '/* End of REXX RSP footer information, compilation time=' (time('e')-startTime) ' */'
Return

/*----------------------------------------------------------------------------*/
/* send message to the console (stdout)                                       */
/*----------------------------------------------------------------------------*/

console_msg: 
If errmsg = 1 Then Say Arg(1)
Return

DSQ: Procedure
line = Arg(1)
x = Pos("'", line)
Do While x > 0
   line = SubStr(line, 1, x)"'"SubStr(line, x + 1)
   x = Pos("'", line, x + 2)
End
Return line

DDQ: Procedure
line = Arg(1)
x = Pos('"', line)
Do While x > 0
   line = SubStr(line, 1, x)'"'SubStr(line, x + 1)
   x = Pos('"', line, x + 2)
End
Return line
